@@ -1,6 +1,6 @@
 /*
     electricsheep - collaborative screensaver
-    Copyright (C) 1999-2001 Scott Draves <source@electricsheep.org>
+    Copyright (C) 1999-2006 Scott Draves <source@electricsheep.org>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
 */
 
 static char *electricsheep_c_id =
-"@(#) $Id: electricsheep.c,v 1.59 2006/04/14 08:59:39 spotspot Exp $";
+"@(#) $Id: electricsheep.c,v 1.67 2006/07/25 20:45:56 spotspot Exp $";
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -77,9 +77,6 @@ int init_delay = 600;
 int init_delay_list = 60;
 int list_freshness = 600;
 
-/* ick but ok since only 5 digits per field in sheep mpg filenames */
-#define max_anims 100000
-
 #define copy_buffer_size 64000
 
 int generation = -1;
@@ -128,7 +125,7 @@ int history_size = 300;
 int *path_history;
 int last_sheep = -1;
 int nrepeated = 0;
-int max_repeats = 3;
+int max_repeats = 2;
 
 
 // by default xscreensaver starts us at priority 10.
@@ -153,6 +150,7 @@ int nobg = 0;
 int parasite = 0;
 int standalone = 0;
 int read_only = 0;
+int show_errors = 1;
 
 int display_zoomed = 0;
 int use_mplayer = 0;
@@ -174,6 +172,8 @@ typedef struct {
     int generation;
     int id;
     int deleted;
+    int readonly;
+    int checked;
     int type;
     time_t ctime;
     int size;
@@ -259,8 +259,9 @@ mysystem(char *cmd, char *msg) {
     if (0) fprintf(stderr, "subprocess; (%s)\n", cmd);
     if (0 != (n = interruptable_system(cmd))) {
 	if (SIGINT != n) {
-	    fprintf(stderr, "subprocess error: %s, %d=%d<<8+%d\n",
-		    msg, n, n>>8, n&255);
+	    if (show_errors)
+		fprintf(stderr, "subprocess error: %s, %d=%d<<8+%d\n",
+			msg, n, n>>8, n&255);
 	    return 1;
 	}
 	fprintf(stderr, "control-c during %s, exiting\n", msg);
@@ -312,12 +313,13 @@ print_anims(char *name, anim_t *an, int nanims) {
     int i;
     printf("%s=%d\n", name, nanims);
     for (i = 0; i < nanims; i++) {
-	printf("gen=%d id=%d deleted=%d type=%d ctime=%d size=%d rating=%d "
+	printf("gen=%d id=%d deleted=%d readonly=%d checked=%d "
+	       "type=%d ctime=%d size=%d rating=%d "
 	       "first=%d last=%d path=%s url=%s\n",
 	       an[i].generation,
-	       an[i].id, an[i].deleted, an[i].type, an[i].ctime,
-	       an[i].size, an[i].rating, an[i].first, an[i].last,
-	       an[i].path, an[i].url);
+	       an[i].id, an[i].deleted, an[i].readonly, an[i].checked, an[i].type,
+	       an[i].ctime, an[i].size, an[i].rating, an[i].first,
+	       an[i].last, an[i].path, an[i].url);
     }
 }
 
@@ -372,10 +374,47 @@ void clear_path(char *path, int generation) {
     closedir(d);
 }
 
+#define bt_zero_block_size (1 << 10)
+
+int bt_zero_blocks(char *fname, int size) {
+    int fd, i;
+    int piece_size = 1 << 18;
+    char buf[bt_zero_block_size];
+
+    if (-1 == (fd = open(fname, O_RDONLY))) {
+	perror(fname);
+	return 1;
+    }
+
+    while (size >= bt_zero_block_size) {
+	int nread;
+	if (bt_zero_block_size != (nread = read(fd, buf, bt_zero_block_size))) {
+	    perror(fname);
+	    return 1;
+	}
+	for (i = 0; i < bt_zero_block_size; i++)
+	    if (buf[i] != 0) break;
+	if (i == bt_zero_block_size)
+	    return 1;
+	if (size < (piece_size + bt_zero_block_size))
+	    break;
+	if (-1 == lseek(fd, piece_size-bt_zero_block_size, SEEK_CUR)) {
+	    perror(fname);
+	    return 1;
+	}
+	size -= piece_size;
+    }
+    return 0;
+}
+
+    
+
 void update_cached_anims_path(char *path, int match_gen) {
     int write_gen = 0;
     int checked_gen = 0;
     int clear_gen = 0;
+    int finished = 0;
+    int ncached_anims_saved = ncached_anims;
     DIR *d;
     struct dirent *e;
     struct stat sbuf;
@@ -405,6 +444,9 @@ void update_cached_anims_path(char *path, int match_gen) {
 	    }
 	}
 	if (e->d_name[0] == '.') continue;
+	if (!strcmp("finished", e->d_name)) {
+	    finished = 1;
+	}
 	if (filename_is_xxx(e->d_name)) {
 	    if (4 != sscanf(e->d_name, "%d=%d=%d=%d.xxx",
 			    &an->generation, &an->id,
@@ -467,6 +509,9 @@ void update_cached_anims_path(char *path, int match_gen) {
 	}
 
 	an->deleted = 0;
+	an->checked = !strcmp(path, leave_prefix);
+	an->readonly = !((sbuf.st_mode & S_IWUSR) &&
+			 (sbuf.st_uid == getuid()));
 	an->size = sbuf.st_size;
 	an->ctime = sbuf.st_ctime;
 	an->url[0] = 0;
@@ -477,6 +522,15 @@ void update_cached_anims_path(char *path, int match_gen) {
     }
     closedir(d);
 
+    if (finished) {
+	int i;
+	for (i = ncached_anims_saved; i < ncached_anims; i++) {
+	    if (!strcmp(cached_anims[i].path, path)) {
+		cached_anims[i].checked = 1;
+	    }
+	}
+    }
+
     if (match_gen && !checked_gen) write_gen = 1;
 
     if (clear_gen) clear_path(path, old_gen);
@@ -532,6 +586,82 @@ print_stats() {
     printf("ncached_anims = %d\n", ncached_anims);
 }
 
+int check_for_eddy() {
+    int i, j, c = 0, n = history_size;
+    if (n > 50) n = 50;
+    for (i = 0; i < n; i++) {
+	int diff = 1;
+	if (-1 == path_history[i]) break;
+	if (debug > 1)
+	    printf("edd checking %d %d %d\n", c, i, path_history[i]);
+	for (j = 0; j < i; j++) {
+	    if (path_history[j] == path_history[i])
+		diff = 0;
+	}
+	c += diff;
+	if (c <= i/3) {
+	    if (debug) printf("eddy %d/%d\n", c, i);
+	    return 1;
+	}
+    }
+    return 0;
+}
+
+int touch(char *fname) {
+    FILE *fp = fopen(fname, "w");
+    if (NULL == fp) {
+	return 1;
+    }
+    fclose(fp);
+    return 0;
+}
+
+
+int check_sheep(int idx) {
+    char fbuf[MAXBUF];
+    int j;
+    int zeroes = 0;
+
+    if (cached_anims[idx].checked) {
+	return 0;
+    }
+    /* should search locally until mismatch */
+    for (j = 0; j < ncached_anims; j++) {
+	
+	if (!strcmp(cached_anims[idx].path,
+		    cached_anims[j].path)) {
+	    cached_file_name(fbuf, &cached_anims[j]);
+	    if (bt_zero_blocks(fbuf, cached_anims[j].size)) {
+		if (debug)
+		    printf("zero_blocks %s\n", fbuf);
+		zeroes = 1;
+		break;
+	    }
+	}
+    }
+    if (!zeroes) {
+	char pbuf[MAXBUF];
+	snprintf(pbuf, MAXBUF, "%sfinished", cached_anims[idx].path);
+	if (debug)
+	    printf("marking finished %s\n", pbuf);
+	touch(pbuf);
+    } else if (debug)
+	printf("unfinished %s\n", cached_anims[idx].path);
+    for (j = 0; j < ncached_anims; j++) {
+	if (!strcmp(cached_anims[idx].path,
+		    cached_anims[j].path)) {
+	    if (zeroes)
+		cached_anims[j].deleted = 1;
+	    else
+		cached_anims[j].checked = 1;
+	}
+    }
+    return zeroes;
+}
+
+int *succs = NULL, *lsuccs = NULL;
+int succs_len, lsuccs_len;
+
 /* traverse the graph of anims, and writes the mpeg files into stdin
    of the child.  another process decodes the mpeg and displays it on
    the screen. */
@@ -542,6 +672,8 @@ do_display() {
     int idx;
     int niters;
     static patience=1;
+    char fbuf[MAXBUF];
+
 
     if (start_id >= 0) {
 	current_anim_id = start_id;
@@ -559,11 +691,13 @@ do_display() {
 	}
     }
 
-    if (-1 == idx) {
-        int n = 0;
-        for (i = 0; i < ncached_anims; i++) {
+    if (idx != -1 && check_sheep(idx)) idx = -1;
+
+    while (-1 == idx) {
+	int n = 0;
+	for (i = 0; i < ncached_anims; i++) {
 	    if (!cached_anims[i].deleted) n++;
-        }
+	}
 	if (0 == n) {
 	    if (patience)
 		fprintf(stderr, "please be patient while the "
@@ -574,29 +708,33 @@ do_display() {
 	    update_cached_anims(0);
 	    return;
 	}
-	idx = irandom(ncached_anims);
-	while (cached_anims[idx].deleted) {
-	    idx = (idx+1)%ncached_anims;
-	}
+	do {
+	    idx = irandom(ncached_anims);
+	} while (cached_anims[idx].deleted);
 	
 	current_anim_id = cached_anims[idx].id;
 	if (debug) printf("found new anim id=%d.\n", current_anim_id);
+
+	if (check_sheep(idx)) idx = -1;
     }
+
+
     if (cached_anims[idx].first == cached_anims[idx].last) {
 	niters = nrepeats;
     } else {
 	niters = 1;
     }
 
+
+
     if (1) {
 	// play an anim
 	int h, i;
 	char pbuf[MAXBUF];
-	char fbuf[MAXBUF];
 	FILE *idf;
 
-	if (debug) printf("play anim x=%d id=%d iters=%d.\n",
-			  idx, current_anim_id, niters);
+	if (debug) printf("play anim x=%d id=%d iters=%d path=%s.\n",
+			  idx, current_anim_id, niters, cached_anims[idx].path);
 	if (stats) print_stats();
 
 	nplays++;
@@ -675,21 +813,25 @@ do_display() {
     else
       nrepeated = 0;
     last_sheep = current_anim_id;
-    
-    if (0 == reset_fuse-- || nrepeated >= max_repeats) {
+
+    if (0 == reset_fuse-- || nrepeated >= max_repeats || check_for_eddy()) {
       if (debug) printf("reset nrepeated=%d reset_fuse=%d\n", nrepeated, reset_fuse);
       current_anim_id = -1;
       reset_fuse = reset_fuse_length;
     } else {
 	// pick next anim at random from all possible succs,
 	// trying to avoid repeats, and giving priority to loops.
-	int succs[max_anims];
-	int lsuccs[max_anims];
 	int nsuccs = 0;
 	int lnsuccs = 0;
 	int sym = cached_anims[idx].last;
 	int h;
-    
+
+	if (NULL == succs) {
+	    succs = malloc(sizeof(int));
+	    succs_len = 1;
+	    lsuccs = malloc(sizeof(int));
+	    lsuccs_len = 1;
+	}
 
 	for (h = history_size; h >= 0; h--) {
 	    for (i = 0; i < ncached_anims; i++) {
@@ -704,10 +846,19 @@ do_display() {
 		    if (hh == h) {
 			if (debug > 1) printf("succ x=%d id=%d h=%d\n",
 					      i, cached_anims[i].id, h);
-			if (cached_anims[i].first == cached_anims[i].last)
+			if (cached_anims[i].first == cached_anims[i].last) {
 			    lsuccs[lnsuccs++] = i;
-			else
+			    if (lnsuccs == lsuccs_len) {
+				lsuccs = realloc(lsuccs, (1+lsuccs_len)*sizeof(int));
+				lsuccs_len++;
+			    }
+			} else {
 			    succs[nsuccs++] = i;
+			    if (nsuccs == succs_len) {
+				succs = realloc(succs, (1+succs_len)*sizeof(int));
+				succs_len++;
+			    }
+			}
 		    }
 		}
 	    }
@@ -767,8 +918,9 @@ delete_cached(int size) {
 	best=0;
 
 	for (i = 0; i < ncached_anims; i++) {
-	    if (cached_anims[i].deleted) continue;
-	    if (0 && debug)
+	    if (cached_anims[i].deleted ||
+		cached_anims[i].readonly) continue;
+	    if (debug)
 	      printf("delete cached total=%g id=%d rating=%d\n",
 		     total, cached_anims[i].id, cached_anims[i].rating);
     
@@ -786,7 +938,10 @@ delete_cached(int size) {
 	    int fd;
 	    cached_file_name(buf, &cached_anims[best]);
 	    if (debug) printf("deleting %s\n", buf);
-	    unlink(buf);
+	    if (-1 == unlink(buf)) {
+		perror(buf);
+		continue;
+	    }
 	    cached_anims[best].deleted = 1;
 	    deleted_file_name(buf, &cached_anims[best]);
 	    if (-1 == (fd = creat(buf, S_IRUSR|S_IWUSR))) {
@@ -817,16 +972,17 @@ download_anim(int idx) {
     snprintf(tfb, MAXBUF, "%s.tmp", cfb);
     strcpy(mpg_name, tfb);
     
-    snprintf(pbuf, MAXBUF, "%s --silent --show-error --output %s %s",
-	    curl_cmd, tfb, server_anims[idx].url);
+    snprintf(pbuf, MAXBUF, "%s --output %s %s",
+	     curl_cmd, tfb, server_anims[idx].url);
 
     if (debug) printf("about to %s\n", pbuf);
 
     mysystem(pbuf, "anim download");
 
     if (-1 == stat(tfb, &sbuf)) {
-	fprintf(stderr, "download failed of sheep %d\n",
-		server_anims[idx].id);
+	if (show_errors)
+	    fprintf(stderr, "download failed of sheep %d\n",
+		    server_anims[idx].id);
 	unlink(tfb);
 	mpg_name[0] = 0;
 	sleep(tryagain);
@@ -861,7 +1017,7 @@ get_control_points(char *buf, int buf_size) {
     char pbuf[MAXBUF];
     FILE *cp;
   
-    snprintf(pbuf, MAXBUF, "%s --silent --show-error 'http://%s/cgi/get?"
+    snprintf(pbuf, MAXBUF, "%s 'http://%s/cgi/get?"
 	    "n=%s&w=%s&v=%s&u=%s' | gunzip -c %s",
 	    curl_cmd, dream_server, nick_buf,
 	    url_buf, client_version, uniqueid, hide_stderr);
@@ -907,7 +1063,7 @@ put_image(char *fname, int frame, int anim_id, int gen) {
 	cleanup_and_exit(1);
     }
 
-    snprintf(pbuf, MAXBUF, "%s --silent --upload-file %s "
+    snprintf(pbuf, MAXBUF, "%s --upload-file %s "
 	    "'http://%s/cgi/put?f=%d&id=%d&s=%ld&g=%d&v=%s&u=%s'",
 	    curl_cmd, fname, dream_server, frame, anim_id,
 	    (long)sbuf.st_size, gen, client_version, uniqueid);
@@ -1019,9 +1175,10 @@ do_render() {
     get_control_points(anim_text, max_cp_size);
 
     if (0 == anim_text[0]) {
-	fprintf(stderr,
-		"lost contact with %s, cannot render frames.\n",
-		dream_server);
+	if (show_errors)
+	    fprintf(stderr,
+		    "lost contact with %s, cannot render frames.\n",
+		    dream_server);
 	sleep(tryagain);
 	return;
     }
@@ -1042,7 +1199,8 @@ do_render() {
     }
     XML_ParserFree(parser);
     if (server_error) {
-	fprintf(stderr, "server reported error for get: %s\n", server_error_type);
+	fprintf(stderr, "server reported error for get: %s\n",
+		server_error_type);
 	sleep(tryagain);
 	return;
     }
@@ -1129,6 +1287,7 @@ list_start_element(void *userData, const char *name, const char **atts) {
 	}
 	memset(an, 0, sizeof(an));
 	an->generation = generation;
+	strncpy(an->path, leave_prefix, PATH_MAX);
 	while (atts[i]) {
 	    const char *a = atts[i+1];
 	    if (!strcmp(atts[i], "id")) {
@@ -1154,12 +1313,8 @@ list_start_element(void *userData, const char *name, const char **atts) {
 	    i += 2;
 	}
 	if (!strcmp(state, "done") && (0 == an->type)) {
-	    if (max_anims == nserver_anims+1) {
-		fprintf(stderr, "too many sheep.\n");
-	    } else {
-		nserver_anims++;
-		server_anims = realloc(server_anims, (1+nserver_anims) * sizeof(anim_t));
-	    }
+	    nserver_anims++;
+	    server_anims = realloc(server_anims, (1+nserver_anims) * sizeof(anim_t));
 	} else if (!strcmp(state, "expunge")) {
 	    char buf[MAXBUF];
 	    if (debug) printf("expunging id=%d.\n", an->id);
@@ -1205,7 +1360,7 @@ void update_server_anims() {
     in_message = 0;
     server_error = 0;
     
-    snprintf(pbuf, MAXBUF, "%s --silent --show-error 'http://%s/cgi/list?v=%s&u=%s'"
+    snprintf(pbuf, MAXBUF, "%s 'http://%s/cgi/list?v=%s&u=%s'"
 	    "| gunzip -c %s", curl_cmd, dream_server,
 	    client_version, uniqueid, hide_stderr);
 
@@ -1251,9 +1406,10 @@ void update_server_anims() {
     }
 
     if (-1 == generation) {
-	fprintf(stderr,
-		"lost contact with %s, cannot retrieve sheep.\n",
-		dream_server);
+	if (show_errors)
+	    fprintf(stderr,
+		    "lost contact with %s, cannot retrieve sheep.\n",
+		    dream_server);
 	sleep(tryagain);
 	return;
     }
@@ -1488,7 +1644,7 @@ make_display_process()  {
 	      while (1) {
 		timestamp();
 		do_display();
-		if (!standalone && ((cnt++%cache_update_delay) == 0)) {
+		if (!standalone && ((++cnt%cache_update_delay) == 0)) {
 		  update_cached_anims(0);
 		}
 	      }
@@ -1660,6 +1816,8 @@ flags_init(int *argc, char ***argv) {
 	else sarg("-window-id", window_id)
 	else iarg("--root", on_root)
 	else iarg("--voting",voting)
+	else iarg("--read-only",read_only)
+	else iarg("--show-errors",show_errors)
 	else iarg("--anim-only",nobg)
 	else iarg("--mplayer",use_mplayer)
 	else iarg("--zoom",display_zoomed)
@@ -1961,6 +2119,10 @@ main(int argc, char **argv) {
 	}
     } else
 	snprintf(curl_cmd, MAXBUF, "nice -n %d curl", nice_level);
+    strcat(curl_cmd, " --silent");
+    if (show_errors) {
+	strcat(curl_cmd, " --show-error");
+    }
 
     srandom(time(0));
 
