@@ -19,11 +19,12 @@
 
 
 static char *jpeg_c_id =
-"@(#) $Id: png.c,v 1.8 2005/07/20 06:06:16 spotspot Exp $";
+"@(#) $Id: png.c,v 1.9 2006/06/26 19:36:49 spotspot Exp $";
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <png.h>
+#include <setjmp.h>
 
 #include "img.h"
 
@@ -67,7 +68,7 @@ unsigned char *read_png(FILE *ifp, int *width, int *height) {
   unsigned char sig_buf [SIG_CHECK_SIZE];
   png_struct *png_ptr;
   png_info *info_ptr;
-  png_byte **png_image;
+  png_byte **png_image = NULL;
   int linesize, x, y;
   unsigned char *p, *q;
 
@@ -85,6 +86,16 @@ unsigned char *read_png(FILE *ifp, int *width, int *height) {
     fprintf (stderr, "cannot allocate LIBPNG structure\n");
     return 0;
   }
+  if (setjmp(png_jmpbuf(png_ptr))) {
+     if (png_image) {
+	 for (y = 0 ; y < info_ptr->height ; y++)
+	     free (png_image[y]);
+	 free (png_image);
+     }
+     png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
+     perror("reading file");
+     return;
+  }
   info_ptr = png_create_info_struct (png_ptr);
   if (info_ptr == NULL) {
     png_destroy_read_struct (&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
@@ -155,9 +166,7 @@ unsigned char *read_png(FILE *ifp, int *width, int *height) {
   for (y = 0 ; y < info_ptr->height ; y++)
     free (png_image[y]);
   free (png_image);
-  png_read_destroy (png_ptr, info_ptr, (png_info *)0);  
-  free (png_ptr);
-  free (info_ptr);
+  png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp)NULL);  
 
   return q;
 }
