@@ -39,6 +39,8 @@ extern int window_id;
 #include <X11/extensions/XShm.h>
 #include <inttypes.h>
 #include <sys/time.h>
+#include <png.h>
+#include <setjmp.h>
 
 #include "vroot.h"
 #include <sys/types.h>
@@ -124,8 +126,6 @@ static void frame_rate_delay()
     tv_rate = tv_end;
 }
 
-#include <png.h>
-
 static char *overlay_luma = NULL;
 static char *overlay_alpha = NULL;
 static int overlay_width;
@@ -165,7 +165,7 @@ static void overlay_read_png(FILE *ifp)
   unsigned char sig_buf [SIG_CHECK_SIZE];
   png_struct *png_ptr;
   png_info *info_ptr;
-  png_byte **png_image;
+  png_byte **png_image = NULL;
   int linesize, x, y;
   unsigned char *p, *q;
 
@@ -183,6 +183,16 @@ static void overlay_read_png(FILE *ifp)
     fprintf (stderr, "cannot allocate LIBPNG structure\n");
     return;
   }
+  if (setjmp(png_jmpbuf(png_ptr))) {
+     if (png_image) {
+	 for (y = 0 ; y < info_ptr->height ; y++)
+	     free (png_image[y]);
+	 free (png_image);
+     }
+     png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
+     perror("reading file");
+     return;
+  }
   info_ptr = png_create_info_struct (png_ptr);
   if (info_ptr == NULL) {
     png_destroy_read_struct (&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
@@ -271,10 +281,7 @@ static void overlay_read_png(FILE *ifp)
   for (y = 0 ; y < info_ptr->height ; y++)
     free (png_image[y]);
   free (png_image);
-  png_read_destroy (png_ptr, info_ptr, (png_info *)0);  
-  free (png_ptr);
-  free (info_ptr);
-
+  png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp)NULL);  
 }
 
 static void overlay_read_ppm(FILE *fin)
@@ -535,6 +542,8 @@ static int open_display (x11_instance_t * instance)
       instance->corner_x = (w - instance->width)/2;
       instance->corner_y = (h - instance->height)/2;
     } else if (window_id == -1) {
+      XTextProperty xname;
+      char *nm = "Electric Sheep";
       /* create a window the same size as the video */
       instance->window =
 	XCreateWindow (instance->display,
@@ -544,6 +553,8 @@ static int open_display (x11_instance_t * instance)
 		       InputOutput, instance->vinfo.visual,
 		       (CWBackPixmap | CWBackingStore | CWBorderPixel |
 			CWEventMask | CWColormap), &attr);
+      XStringListToTextProperty(&nm, 1, &xname);
+      XSetWMName(instance->display, instance->window, &xname);
     } else {
 	/* zoomed to fit the window specified on the command line */
 	XWindowAttributes xgwa;
